''
'' VGA display 80x25 (dual cog) - demo
''
''        Author: Marko Lukat
'' Last modified: 2014/09/16
''       Version: 0.4
''
CON
  _clkmode = XTAL1|PLL16X
  _xinfreq = 5_000_000

CON
  columns  = driver#res_x / 9
  rows     = driver#res_y / font#height
  bcnt     = columns * rows

  rows_raw = (driver#res_y + font#height - 1) / font#height
  bcnt_raw = columns * rows_raw
 ' bcnt_raw = columns * 25
 
  lastrow = rows-1
  lastcolumn = columns-1

CON
  vgrp     = 2                                          ' video pin group
  escape = 27
  bracket = 91
  semicolon = 59
  maxparam = 7    ' maximum numbers of params 'on stack' - 1 we can take for ESC [ n1; n2; n3 .. n_maxparam-1 ..
  
OBJ
  serial: "FullDuplexSerial"
  driver: "waitvid.80x25.driver.2048"
    font: "halfrange8x16-2font"
  
VAR
  long  scrn[bcnt_raw / 2]                              ' screen buffer
  long  link[driver#res_m]                              ' mailbox
  byte  attribute
  long col, row

  long rxChar
  byte ecm_state           ' state of the ECMA state machine
  word param_n[maxparam+1] ' ESC [ n1;n2 ... nx .. parameters
  word param_ptr           ' poitner and count-1 params on stack
  word param_r             ' parameter read pointer
  word param_flag
 
PUB selftest : n | c, aval

  link{0} := constant((vgrp << 9 | %%333_3) << 21) | @scrn{0}
  link[1] := font#height << 24 | font.addr
  
  driver.init(-1, @link{0})                             ' start driver

' start the serial driver too
   serial.Start(6, 7, %0000, 9_600)                     ' requires 1 cog for operation
  
' Initial attribute black background, white colour

  attribute := %11110000

  ecm_state := 0
  param_ptr := param_r := param_flag := 0

' Fill all the video with space

  c := 32

'   colour format: %FFFF_BBB_A
'
'     FFFF: foreground index
'      BBB: background index


  col := 0
  row := 0

  aval := c << 8 | attribute
  wordfill (@scrn, aval, bcnt )

  
  at (40,12)
  str ( string (" This is a test") )
  setAttribute ( 1,0,0 )
  at (0,0)
  str ( string ("Writing in different colour @(0,13) ") )
  setAttribute ( 2,0,0 )
  at (0,24)
  str ( string ("This is the last line") )
  at (0,1)
  
repeat
 rxChar :=  serial.RxCheck 
  if rxChar > 0
   print ( rxChar )
while 1
  
PUB clearScreen | c, aval

  c := 32
  aval := c << 8 | attribute
  col := 0
  row := 0
  wordfill (@scrn, aval, bcnt )

PUB at (x,y)
 col := x
 row := y

PUB setAttribute ( foreground, background, flash )
 attribute := ( foreground & 15 ) << 4 | (background & 7 ) << 1 | flash & 1
 
PUB str(stringptr)

'' Print a zero-terminated string

  repeat strsize(stringptr)
    print(byte[stringptr++])
  
PRI redef(c, cdef) : s

  repeat s from 0 to 15 step 2
    word[font.addr][64 * s + c] := byte[cdef][s] << 8 | byte[cdef][s+1]

PRI newline | c, aval, y
'
  col := 0
  c := 32
  aval := c << 8 | attribute
  
  if ++row == rows
    y := 24
    repeat rows-1
     wordmove(@scrn+160*y, @scrn+160*y-160, 80)   'scroll lines
     y := y-1
    wordfill(@scrn, aval, 80) 'fill last line
    row := rows-1
 

PRI printAt(x, y, c, attr)

  if x < columns AND x > -1 AND y < rows AND y > -1
   attr.byte[1] := c
   scrn.word[bcnt_raw - y * columns - ++x] := attr


PRI isDigit ( c ) : okay
okay := 0
 if c > 47 AND c < 58
  okay := 1

PRI getPNonZero : n1
n1 := 1
 if param_ptr > 0 
  n1 := param_n[param_r]
  ++param_r
  --param_ptr
  if n1 == 0
   n1 := 1
 

PRI getP : n2
n2 := 0
 if param_ptr > 0
  n2 := param_n[param_r]
  ++param_r
  --param_ptr

PRI clearlin ( mode ) | aval
  aval := " " << 8 | attribute
 if mode == 0
    wordfill(@scrn+(24-row)*160, aval, 80-col) ' from cursor to end line
 if mode == 1
    wordfill(@scrn+(24-row)*160+2*(80-col-1), aval, col+1) ' from startline to cursor
 if mode == 2
    wordfill(@scrn+(24-row)*160, aval, 80) ' clear all line
    
PRI clearscrin ( mode ) | aval
  aval := " " << 8 | attribute
 if mode == 0
    wordfill(@scrn, aval, 80*(24-row)+(80-col)) ' from cursor to end line
 if mode == 1
    wordfill(@scrn+(24-row)*160+2*(80-col-1), aval, 80*row+col+1) ' from cursor to end line
 if mode == 2
    wordfill(@scrn, aval, columns*rows) ' clear all the screen
    col := row := 0
  
PRI scroll ( mode, n ) | aval, y1,y2
  aval := "W" << 8 | attribute
  if mode == 0
    y1 := 24
    y2 := 24-n
    repeat rows-n
     wordmove(@scrn+160*y1, @scrn+160*y2, 80)   'scroll lines
     y1--
     y2--
   y1 := 0
   repeat n   
    wordfill(@scrn+160*y1, aval, 80) 'fill last line
    y1++ 
  
  if mode == 1
    y2 := 24-(rows-n)
    y1 := y2-1
    repeat rows-n
     wordmove(@scrn+160*y2, @scrn+160*y1, 80)   'scroll lines
     y1++
     y2++
   y1 := 24
   repeat n   
    wordfill(@scrn+160*y1, aval, 80) 'fill last line
    y1-- 

PRI doCases(c) | i , n1 , n2
 case c
    ' cursor UP n1 lines
  "A" :  
   n1 := getPNonZero
   row := row-n1
   if (row < 0)
    row := 0

  ' cursor DOWN n1 lines
  "B" :
   n1 := getPNonZero
   row := row+n1
   if (row > lastrow)
       row := lastrow

  ' cursor RIGHT n1 lines
  "C" :
   n1 := getPNonZero
   col := col+n1
   if (col > lastcolumn)
        col := lastcolumn

  ' cursor LEFT n1 lines
  "D" :
   n1 := getPNonZero
   col := col-n1
   if ( col < 0 )
        col := 0

  ' cursor NEXT n1 lines
  "E" :
   n1 := getPNonZero
   row := row+n1
   if ( row > lastrow )
        row := lastrow
   col := 0
   
  ' cursor PREV n1 lines
  "F" :
   n1 := getPNonZero
   row := row-n1
   if ( row < 0 )
       row := 0
   col := 0

  ' cursor Horizontal absolute
  "G" :
   n1 := getPNonZero
   col := n1-1
   if ( col > lastcolumn )
       col := lastcolumn

  ' cursor Position absolute
  "H","f" :
   n1 := getPNonZero
   n2 := getPNonZero
   row := n1-1
   if ( row > lastrow )
       row := lastrow
   col := n2-1
   if ( col > lastcolumn )
       col := lastcolumn

  ' clear screen
  "J" :
   n1 := getP
   clearscrin (n1)
   
  ' clear line
  "K" :
   n1 := getP
   clearlin (n1)

  ' scroll up
  "S" :
   n1 := getPnonZero
   if n1 > rows
    n1 := rows
   scroll (0,n1)

  ' scroll down
  "T" :
   n1 := getPnonZero
   if n1 > rows
    n1 := rows
   scroll (1,n1)
      
  OTHER :
   param_ptr := 0
   param_flag := param_r := 0
   ecm_state := 0
   
       
PRI print (c) | i , n1, n2

if ecm_state == 0

 ' carriage return
 if c == $0d
  col := 0
  return

 ' line feed 
 if c == $0a
  newline
  return

 ' backspace 
 if c == $08
  col := col-1
  if col < 0
   col := 0
  return

 ' first stage we got an escape
 if c == escape
  ecm_state := 1                
  return

 ' print whatever it is, we are in state 0  
 printAt (col, row, c, attribute)

 if ++col == columns
  newline
 return
  
' second stage, we got and ESC [
if ecm_state == 1
 param_ptr := param_r := param_flag := 0 ' reset stack ptr
 if c == bracket
  ecm_state := 2
  repeat i from 0 to maxparam ' reset initial values
   param_n[i] := 0
  return
 else 
  ' we got an escape followed by something totally else, ignore
   ecm_state := 0
   return

' we got an ESC [, now we begin collect number Nx until ; or else 
if ecm_state == 2 
  if c <> semicolon
   if isDigit(c)
    param_n[param_ptr] := param_n[param_ptr]*10+c-48
    param_flag := 1
    return
   else
    ecm_state := 4 ' c is still a valid char to interpret
    if param_flag > 0
     ++param_ptr
    if param_ptr > maxparam
     param_ptr := maxparam 
    param_flag := 0
  else
   ++param_ptr
   if param_ptr > maxparam
    param_ptr := maxparam
   return
   
if ecm_state == 4

 repeat 
  doCases(c)
 while param_ptr > 0
 
 param_ptr := param_r := param_flag := 0       
 ecm_state := 0 ' back to initial state
   
DAT